from __future__ import print_function, division, absolute_import, unicode_literals
import collections

import six
from six.moves import zip
from past.builtins import basestring
from .utils.frozen_dict import OrderedFrozenDict, SortedFrozenDict


class DataDefinition(OrderedFrozenDict):
    def __init__(self, key, args=None, name=None):
        assert isinstance(key, (basestring, Argument, tuple)), \
            "Data key can only be str or Argument."
        assert name is None or isinstance(name, basestring), "Data name can only be str."
        self._key = key
        if args is None:
            self._args = SortedFrozenDict()
        else:
            self._args = SortedFrozenDict.recursively_froze(args)
        self._name = name
        super(DataDefinition, self).__init__((('key', key), ('args', self._args)))

    @property
    def key(self):
        return self._key

    @property
    def args(self):
        return self._args

    @property
    def name(self):
        return self._name

    def replace(self, key=None, args=None, name=None):
        if key is None:
            key = self._key
        if args is None:
            args = self._args
        if name is None:
            name = self._name
        return DataDefinition(key=key, args=args, name=name)

    def __str__(self):
        class_name = type(self).__name__
        if len(self._args) == 0:
            return "%s(%r)" % (class_name, self._key)
        return "%s(%r, %s)" % (class_name, self._key, self._args)

    def __repr__(self):
        return str(self)


class Argument(object):
    """Represent how the argument is passed from downstream to upstream.

    Parameters
    ----------
    parameter : str
        The name of the parameter from downstream data definition.
    callable : Optional[Callable]
        The callable to be applied on the downstream argument. If None, use identity function.
    """

    def __init__(self, parameter, callable=None):
        self.parameter = parameter
        self.callable = callable

    def eval(self, args):
        if self.callable is None:
            return args[self.parameter]
        return self.callable(args[self.parameter])

    def __str__(self):
        class_name = type(self).__name__
        return "%s(%r)" % (class_name, self.parameter)

    def __repr__(self):
        return str(self)

    def __gt__(self, other):
        if isinstance(other, basestring):
            return self.parameter > other
        elif isinstance(other, Argument):
            return self.parameter > other.parameter
        return NotImplemented

    def __lt__(self, other):
        if isinstance(other, basestring):
            return self.parameter < other
        elif isinstance(other, Argument):
            return self.parameter < other.parameter
        return NotImplemented


class RequirementDefinition(DataDefinition):
    def eval_data_definition(self, args):
        # TODO: refactor
        # evaluate key
        if isinstance(self._key, Argument):
            raw_data_def = self._key.eval(args)
            if isinstance(raw_data_def, collections.Mapping):
                new_keys = [raw_data_def['key']]
                new_args = [dict(raw_data_def.get('args', {}))]
            elif isinstance(raw_data_def, basestring):
                new_keys = [raw_data_def]
                new_args = [{}]
            elif isinstance(raw_data_def, collections.Sequence):
                new_keys = []
                new_args = []
                for _raw_data_def in raw_data_def:
                    if isinstance(_raw_data_def, basestring):
                        new_key = _raw_data_def
                        new_arg = {}
                    else:
                        new_key = _raw_data_def['key']
                        new_arg = dict(_raw_data_def.get('args', {}))
                    new_keys.append(new_key)
                    new_args.append(new_arg)
            else:
                raise ValueError("Evaluated arguments only support dict, list or str.")
        elif isinstance(self._key, basestring):
            new_keys = [self._key.format(**args)]
            new_args = [{}]
        else:
            raise ValueError("RequirementDefinition.key can only be Argument or str.")

        # evaluate arguments
        for key, arg in six.viewitems(self._args._dict):
            if isinstance(arg, Argument):
                arg = arg.eval(args)
            elif isinstance(arg, basestring):
                arg = arg.format(**args)
            elif not isinstance(arg, collections.Hashable):
                raise ValueError(
                    "The values in RequirementDefinition.args can only be Argument or hashable.")
            for new_arg in new_args:
                new_arg[key] = arg

        data_definitions = [DataDefinition(key, arg, self._name)
                            for key, arg in zip(new_keys, new_args)]
        return data_definitions



from __future__ import print_function, division, absolute_import, unicode_literals
from os.path import dirname

from mkdir_p import mkdir_p
import six
import networkx as nx


def draw_dag(nx_dag, path):
    if dirname(path) != '':
        mkdir_p(dirname(path))
    agraph = nx.nx_agraph.to_agraph(nx_dag)
    for edge in agraph.edges_iter():
        if edge.attr['nonskipped_data'] is None:
            edge.attr['label'] = edge.attr['data_definitions']
        else:
            edge.attr['label'] = ""
            if edge.attr['nonskipped_data'] not in ["set()", "set([])"]:
                edge.attr['label'] += edge.attr['nonskipped_data']
            if (edge.attr['skipped_data'] not in ["set()", "set([])"]
                    and edge.attr['skipped_data'] is not None):
                edge.attr['label'] += "(%s skipped)" % edge.attr['skipped_data']
    for node in agraph.nodes_iter():
        if node.attr['skipped'] == "True":
            node.attr['label'] = node.attr['func_name'] + " (skipped)"
            node.attr['fontcolor'] = 'grey'
        else:
            node.attr['label'] = node.attr['func_name']
    agraph.layout('dot')
    agraph.draw(path)


class DataGraph(object):
    """Data that represents the dependency.
    """

    def __init__(self):
        self._key_output_config_dict = {}
        self._key_node_attrs_dict = {}

    def add_node(self, name, parameters, requirements, output_configs):
        # pylint: disable=protected-access
        # format better data structure
        parameters = tuple(parameters)
        requirements = tuple(requirements)
        output_config_dict = {
            config['key']: {
                'handler': config['handler'],
                'handler_kwargs': config['handler_kwargs'],
            }
            for config in output_configs
        }
        if not output_config_dict:
            raise ValueError("No data key for {}.".format(name))

        # build index
        for key, output_config in six.viewitems(output_config_dict):
            if key in self._key_output_config_dict:
                raise ValueError("Duplicated data key '{}' in {} (alreadly have value {})."
                                 .format(key, name, self._key_output_config_dict[key]))
            self._key_output_config_dict[key] = output_config

        node_attrs = {
            'func_name': name,
            'parameters': parameters,
            'requirements': requirements,
            'output_configs': output_config_dict,
        }

        for key in output_config_dict.keys():
            if key in self._key_node_attrs_dict:
                raise ValueError("Duplicated data key '{}' for {} and {}."
                                 .format(key, self._key_node_attrs_dict[key], node_attrs))
            self._key_node_attrs_dict[key] = node_attrs

    def get_handler_name(self, key):
        return self._key_output_config_dict[key]['handler']

    def _grow_ancestors(self, nx_digraph, root_node_key, predecessor_defs):
        """
        Parameters
        ----------
        predecessor_defs: Sequence[DataDefinition]
        """
        # grow the graph using DFS
        for predecessor_def in predecessor_defs:
            node_attrs = self._key_node_attrs_dict[predecessor_def.key]

            # for merging node, we use key as the 'key' in nx_digraph
            node_data_defs = predecessor_def.replace(
                key=tuple(node_attrs['output_configs'].keys()))

            if node_data_defs not in nx_digraph:
                # ancestors of this node has not been grown
                # TODO: check the argument
                requirement_defs = [data_def
                                    for req in node_attrs['requirements']
                                    for data_def in req.eval_data_definition(predecessor_def.args)]
                data_name_set = {req_def.name for req_def in requirement_defs}

                # check duplicated data name
                if len(node_attrs['requirements']) != len(data_name_set):
                    raise ValueError("Duplicated data names exist: {}".format(requirement_defs))
                nx_digraph.add_node(node_data_defs, **node_attrs)
                self._grow_ancestors(nx_digraph, node_data_defs, requirement_defs)

            if not nx_digraph.has_edge(node_data_defs, root_node_key):
                # initialize edge
                nx_digraph.add_edge(node_data_defs, root_node_key, data_definitions=[])

            # set edge attributes
            edge_attr = nx_digraph.edges[node_data_defs, root_node_key]
            edge_attr['data_definitions'].append(predecessor_def)

    def build_directed_graph(self, data_definitions, root_node_key='root'):
        """
        Parameters
        ----------
        data_definitions: Sequence[DataDefinition]
        """
        nx_digraph = nx.DiGraph()
        nx_digraph.add_node(root_node_key, func_name=root_node_key)
        self._grow_ancestors(nx_digraph, root_node_key, data_definitions)
        return nx_digraph

    def draw(self, path, data_definitions, root_node_key='root', reverse=False):
        nx_digraph = self.build_directed_graph(data_definitions, root_node_key)
        if reverse:
            nx_digraph.reverse(copy=False)
        draw_dag(nx_digraph, path)
        return nx_digraph



from __future__ import print_function, division, absolute_import, unicode_literals
from io import StringIO

import dagian
from dagian import Argument as A
from dagian.decorators import (
    require,
    will_generate,
)
import numpy as np
import pandas as pd
from scipy.sparse import csr_matrix
from sklearn.model_selection import train_test_split


class LifetimeFeatureGenerator(dagian.FeatureGenerator):

    @will_generate('memory', 'data_df')
    def gen_data_df(self):
        csv = StringIO("""\
id,lifetime,tested_age,weight,height,gender,income
0, 68, 50, 60.1, 170.5, f, 22000
1, 59, 41, 90.4, 168.9, m, 19000
2, 52, 39, 46.2, 173.6, m, 70000
3, 68, 25, 93.9, 180.0, m, 1000000
4, 99, 68, 65.7, 157.6, f, 46000
5, 90, 81, 56.3, 170.2, f, 17000
""")
        return {'data_df': pd.read_csv(csv, index_col='id')}

    @require('data_df')
    @will_generate('h5py', 'lifetime')
    def gen_lifetime(self, upstream_data):
        data_df = upstream_data['data_df']
        return {'lifetime': data_df['lifetime']}

    @require('data_df')
    @will_generate('h5py', ['weight', 'height'])
    def gen_raw_data_features(self, upstream_data):
        data_df = upstream_data['data_df']
        return data_df[['weight', 'height']]

    @require('data_df')
    @will_generate('memory', 'mem_raw_data')
    def gen_mem_raw_data(self, upstream_data):
        data_df = upstream_data['data_df']
        return {'mem_raw_data': data_df[['weight', 'height']].values}

    # @require('data_df')
    # @will_generate('h5py', 'man_raw_data', manually_create_dataset=True)
    # def gen_man_raw_data(self, upstream_data, create_dataset_functions):
    #     data_df = upstream_data['data_df']
    #     dset = create_dataset_functions['man_raw_data'](shape=(data_df.shape[0], 2))
    #     dset[...] = data_df[['weight', 'height']].values

    @require('data_df')
    @will_generate('pandas_hdf', ['pd_weight', 'pd_height'])
    def gen_raw_data_table(self, upstream_data):
        data_df = upstream_data['data_df']
        result_df = data_df.loc[:, ['weight', 'height']]
        result_df.rename(columns={'weight': 'pd_weight', 'height': 'pd_height'},
                         inplace=True)
        return result_df

    @require('data_df')
    @will_generate('pandas_hdf', 'pd_raw_data')
    def gen_raw_data_df(self, upstream_data):
        data_df = upstream_data['data_df']
        return {'pd_raw_data': data_df[['weight', 'height']]}

    # @require('pd_raw_data')
    # @will_generate('pandas_hdf', 'pd_raw_data_append', manually_append=True)
    # def gen_raw_data_append_df(self, upstream_data, append_functions):
    #     df = upstream_data['pd_raw_data'].value
    #     append_functions['pd_raw_data_append'](df.iloc[:3])
    #     append_functions['pd_raw_data_append'](df.iloc[3:])

    @require('data_df')
    @will_generate('h5py', 'BMI')
    def gen_bmi(self, upstream_data):
        data_df = upstream_data['data_df']
        bmi = data_df['weight'] / ((data_df['height'] / 100) ** 2)
        return {'BMI': bmi}

    @require('{dividend}')
    @require('{divisor}')
    @will_generate('h5py', 'division')
    def gen_division(self, upstream_data, dividend, divisor='height'):
        division_result = upstream_data['{dividend}'].value / upstream_data['{divisor}'].value
        return {'division': division_result}

    @require('division', 'partial_division', dividend=A('dividend'), divisor=A('divisor1'))
    @require('{divisor2}', 'divisor2')
    @will_generate('h5py', 'division_2_divisor')
    def gen_division_2_divisor(self, upstream_data, dividend, divisor1, divisor2):
        division_result = upstream_data['partial_division'].value / upstream_data['divisor2'].value
        return {'division_2_divisor': division_result}

    @require('division', dividend=A('dividend', lambda x: 'pd_' + x), divisor=A('divisor1'))
    @require(A('divisor2'))
    @will_generate('h5py', 'division_pd_2_divisor')
    def gen_division_pd_2_divisor(self, upstream_data, dividend, divisor1, divisor2):
        division_result = upstream_data['division'].value / upstream_data['divisor2'].value
        return {'division_pd_2_divisor': division_result}

    @require(A('dividend'))
    @require(A('divisor'))
    @will_generate('h5py', 'recursive_division')
    def gen_recursive_division(self, upstream_data, dividend, divisor):
        division_result = upstream_data['dividend'].value / upstream_data['divisor'].value
        return {'recursive_division': division_result}

    @require(A('sequence'))
    @will_generate('h5py', 'sequential_division')
    def gen_sequential_division(self, upstream_data, sequence):
        assert sequence
        division_result = upstream_data['sequence'][0].value
        for data in upstream_data['sequence'][1:]:
            division_result /= data.value
        return {'sequential_division': division_result}

    @require('data_df')
    @will_generate('pickle', 'train_test_split')
    def gen_train_test_split(self, upstream_data):
        data_df = upstream_data['data_df']
        train_id, test_id = train_test_split(
            data_df.index, test_size=0.5, random_state=0)
        return {'train_test_split': (train_id, test_id)}

    @require('data_df')
    @require('train_test_split')
    @will_generate('h5py', 'is_in_test_set')
    def gen_is_in_test_set(self, upstream_data):
        data_df = upstream_data['data_df']
        _, test_id = upstream_data['train_test_split']
        is_in_test_set = data_df.index.isin(test_id)
        sparse_is_in_test_set = csr_matrix(is_in_test_set[:, np.newaxis])
        return {'is_in_test_set': sparse_is_in_test_set}

    @require('data_df')
    @will_generate('h5py', 'nan', allow_nan=True)
    def gen_nan(self, upstream_data):
        nan = np.full(upstream_data['data_df'].shape[0], fill_value=np.nan, dtype=np.float32)
        return {'nan': nan}



from __future__ import print_function, division, absolute_import, unicode_literals
import unittest
from collections import OrderedDict
import json

import six

from dagian.utils.frozen_dict import FrozenDict, SortedFrozenDict, OrderedFrozenDict


class FrozenDictTest(unittest.TestCase):
    def setUp(self):
        self.original_dict = {'123': '456', '789': '000'}
        self.frozen_dict = FrozenDict(self.original_dict)

    def test_getitem(self):
        for key, val in six.viewitems(self.original_dict):
            self.assertEqual(val, self.frozen_dict[key])

    def test_contains(self):
        self.assertEqual('a' in self.original_dict, 'a' in self.frozen_dict)
        self.assertEqual('123' in self.original_dict, '123' in self.frozen_dict)
        self.assertEqual('789' in self.original_dict, '789' in self.frozen_dict)
        self.assertEqual(789 in self.original_dict, 789 in self.frozen_dict)

    def test_iter(self):
        original_keys = sorted(self.original_dict)
        frozen_keys = sorted(self.frozen_dict)
        self.assertListEqual(original_keys, frozen_keys)

    def test_len(self):
        self.assertEqual(len(self.original_dict), len(self.frozen_dict))

    def test_str(self):
        self.assertIsInstance(str(self.frozen_dict), str)

    def test_repr(self):
        self.assertEqual('FrozenDict(%s)' % repr(self.frozen_dict._dict), repr(self.frozen_dict))

    def test_hash(self):
        self.assertIsInstance(hash(self.frozen_dict), int)

    def test_keys(self):
        original_keys = sorted(six.viewkeys(self.original_dict))
        frozen_keys = sorted(self.frozen_dict.keys())
        self.assertListEqual(original_keys, frozen_keys)

    def test_values(self):
        original_values = sorted(six.viewvalues(self.original_dict))
        frozen_values = sorted(self.frozen_dict.values())
        self.assertListEqual(original_values, frozen_values)

    def test_items(self):
        original_items = sorted(six.viewitems(self.original_dict))
        frozen_items = sorted(self.frozen_dict.items())
        self.assertListEqual(original_items, frozen_items)

    def test_copy(self):
        new_frozen_dict = self.frozen_dict.copy()
        self.assertDictEqual(dict(new_frozen_dict), dict(self.frozen_dict))

    def test_replace(self):
        new_frozen_dict = self.frozen_dict.replace({'123': 123}, a=2)
        self.assertDictEqual(dict(self.original_dict), dict(self.frozen_dict))
        self.assertEqual(123, new_frozen_dict['123'])
        self.assertEqual(2, new_frozen_dict['a'])

    def test_to_json(self):
        self.assertDictEqual(self.original_dict, json.loads(self.frozen_dict.to_json()))

    def test_operators(self):
        self.assertTrue(self.frozen_dict == self.frozen_dict)
        self.assertFalse(self.frozen_dict != self.frozen_dict)
        self.assertFalse(self.frozen_dict < self.frozen_dict)
        self.assertTrue(self.frozen_dict <= self.frozen_dict)
        self.assertFalse(self.frozen_dict > self.frozen_dict)
        self.assertTrue(self.frozen_dict >= self.frozen_dict)


class OrderedFrozenDictTest(FrozenDictTest):
    def setUp(self):
        self.original_dict = OrderedDict((('123', '456'), ('789', '000')))
        self.frozen_dict = OrderedFrozenDict(self.original_dict)

    def test_repr(self):
        self.assertEqual('OrderedFrozenDict(%s)' % self.original_dict, repr(self.frozen_dict))

    def test_iter(self):
        original_keys = list(self.original_dict)
        frozen_keys = list(self.frozen_dict)
        self.assertListEqual(original_keys, frozen_keys)

    def test_keys(self):
        original_keys = list(six.viewkeys(self.original_dict))
        frozen_keys = list(self.frozen_dict.keys())
        self.assertListEqual(original_keys, frozen_keys)

    def test_values(self):
        original_values = list(six.viewvalues(self.original_dict))
        frozen_values = list(self.frozen_dict.values())
        self.assertListEqual(original_values, frozen_values)

    def test_items(self):
        original_items = list(six.viewitems(self.original_dict))
        frozen_items = list(self.frozen_dict.items())
        self.assertListEqual(original_items, frozen_items)

    def test_to_json(self):
        self.assertEqual(json.dumps(self.original_dict), self.frozen_dict.to_json())


class SortedFrozenDictTest(OrderedFrozenDictTest):
    def setUp(self):
        self.original_dict = OrderedDict((('123', '456'), ('789', '000')))
        unsorted_dict = OrderedDict((('789', '000'), ('123', '456')))
        self.frozen_dict = SortedFrozenDict(unsorted_dict)

    def test_repr(self):
        self.assertEqual('SortedFrozenDict(%s)' % self.original_dict, repr(self.frozen_dict))



from __future__ import print_function, division, absolute_import, unicode_literals
import sys
import argparse

import yaml

from .config import (
    get_data_generator_class_from_config,
    get_data_generator_from_config,
)
from ..bundling import get_data_definitions_from_structure


def draw_dag(argv=sys.argv[1:]):
    parser = argparse.ArgumentParser(
        description="Generate DAG.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('-g', '--global-config',
                        default=".dagianrc/config.yml",
                        help="the path of the path configuration YAML file")
    parser.add_argument('-b', '--bundle-config',
                        default=".dagianrc/bundle_config.yml",
                        help="the path of the bundle configuration YAML file")
    parser.add_argument('-d', '--dag-output-path',
                        default="dag.svg",
                        help="output image path")
    parser.add_argument('-i', '--involved', action='store_true',
                        help="annotate the involved nodes and skipped nodes")
    args = parser.parse_args(argv)
    with open(args.global_config) as fp:
        global_config = yaml.safe_load(fp)
    with open(args.bundle_config) as fp:
        bundle_config = yaml.safe_load(fp)
    data_definitions = get_data_definitions_from_structure(bundle_config['structure'])
    if args.involved:
        data_generator = get_data_generator_from_config(global_config)
        data_generator.draw_involved_dag(args.dag_output_path, data_definitions)
    else:
        generator_class = get_data_generator_class_from_config(global_config)
        generator_class.draw_dag(args.dag_output_path, data_definitions)



from __future__ import print_function, division, absolute_import, unicode_literals
import os.path
from abc import ABCMeta, abstractmethod
from functools import partial
import warnings

from bistiming import SimpleTimer
import h5py
import h5sparse
from mkdir_p import mkdir_p
import numpy as np
import pandas as pd
import scipy.sparse as ss
import six
from six.moves import cPickle
from tables import NaturalNameWarning
from pathlib2 import Path

from .data_wrappers import PandasHDFDataset
from .data_definition import DataDefinition


SPARSE_FORMAT_SET = set(['csr', 'csc'])


class DataHandler(six.with_metaclass(ABCMeta, object)):

    @abstractmethod
    def can_skip(self, data_definition):
        pass

    @abstractmethod
    def get(self, data_definition):
        pass

    def get_function_kwargs(self, will_generate_keys, data):
        # pylint: disable=unused-argument
        kwargs = {}
        if len(data) > 0:
            kwargs['data'] = data
        return kwargs

    @abstractmethod
    def write_data(self, data_definition, data, **kwargs):
        pass

    def bundle(self, data_definition, path, new_key):
        """Copy the data to another HDF5 file with new key."""
        data = self.get(data_definition)
        with h5sparse.File(path) as h5f:
            h5f.create_dataset(new_key, data=data)
        self.close()

    def is_return_data_expected(self, **kwargs):
        return True

    def close(self):
        pass


class H5pyDataHandler(DataHandler):

    def __init__(self, hdf_path):
        hdf_dir = os.path.dirname(hdf_path)
        if hdf_dir != '':
            mkdir_p(hdf_dir)
        # create an empty file if the file doesn't exists
        h5py.File(hdf_path, 'a').close()
        self.hdf_path = hdf_path
        self.h5f = None

    def can_skip(self, data_definition):
        with h5py.File(self.hdf_path, 'r') as h5f:
            if data_definition.to_json() in h5f:
                return True
        return False

    def get_read_only_h5py_file(self):
        if self.h5f is None:
            self.h5f = h5py.File(self.hdf_path, 'r')
        return self.h5f

    def get(self, data_definition):
        h5f = self.get_read_only_h5py_file()
        if isinstance(data_definition, DataDefinition):
            return h5sparse.Group(h5f)[data_definition.to_json()]
        return {k: h5sparse.Group(h5f)[k.to_json()] for k in data_definition}

    def get_function_kwargs(self, will_generate_keys, data,
                            manually_create_dataset=False):
        kwargs = {}
        if len(data) > 0:
            kwargs['data'] = data
        if manually_create_dataset is True:
            kwargs['create_dataset_functions'] = {
                k: partial(self.h5f.create_dataset, k)
                for k in will_generate_keys
            }
        elif manually_create_dataset in SPARSE_FORMAT_SET:
            kwargs['create_dataset_functions'] = {
                k: partial(h5sparse.Group(self.h5f).create_dataset, k)
                for k in will_generate_keys
            }
        return kwargs

    def write_data(self, data_definition, data, allow_nan=False):
        with h5py.File(self.hdf_path, 'a') as h5f:
            if not allow_nan:
                # check nan
                if ss.isspmatrix(data):
                    if np.isnan(data.data).any():
                        raise ValueError("data {} have nan".format(data_definition))
                elif np.isnan(data).any():
                    raise ValueError("data {} have nan".format(data_definition))

            # write data
            with SimpleTimer("[{}] Writing generated data {} to hdf5 file"
                             .format(type(self).__name__, data_definition),
                             end_in_new_line=False):
                if data_definition.to_json() in h5f:
                    # h5f[key][...] = data
                    raise NotImplementedError(
                        "Overwriting not supported. Please report an issue.")
                else:
                    h5sparse.Group(h5f).create_dataset(data_definition.to_json(), data=data)

    def close(self):
        if self.h5f is not None:
            self.h5f.close()
            self.h5f = None


class PandasHDFDataHandler(DataHandler):

    def __init__(self, hdf_path):
        hdf_dir = os.path.dirname(hdf_path)
        if hdf_dir != '':
            mkdir_p(hdf_dir)
        # create an empty file if the file doesn't exists
        pd.HDFStore(hdf_path, 'a').close()
        self.hdf_path = hdf_path
        self.hdf_store = None

    def can_skip(self, data_definition):
        with pd.HDFStore(self.hdf_path, 'r') as hdf_store:
            if data_definition.to_json() in hdf_store:
                return True
        return False

    def get_read_only_hdf_store(self):
        if self.hdf_store is None:
            self.hdf_store = pd.HDFStore(self.hdf_path, 'r')
        return self.hdf_store

    def get(self, data_definition):
        hdf_store = self.get_read_only_hdf_store()
        if isinstance(data_definition, DataDefinition):
            return PandasHDFDataset(hdf_store, data_definition.to_json())
        return {k: PandasHDFDataset(hdf_store, k.to_json()) for k in data_definition}

    def get_function_kwargs(self, will_generate_keys, data,
                            manually_append=False):
        kwargs = {}
        if len(data) > 0:
            kwargs['data'] = data
        if manually_append is True:
            kwargs['append_functions'] = {
                k: partial(self.hdf_store.append, k)
                for k in will_generate_keys
            }
        return kwargs

    def write_data(self, data_definition, data, allow_nan=False):
        with pd.HDFStore(self.hdf_path, 'a') as hdf_store:
            if not allow_nan:
                # check nan
                is_null = False
                if isinstance(data, pd.DataFrame):
                    if data.isnull().any().any():
                        is_null = True
                elif isinstance(data, pd.Series):
                    if data.isnull().any():
                        is_null = True
                else:
                    raise ValueError("PandasHDFDataHandler doesn't support type {} (in key {})"
                                     .format(type(data), data_definition))
                if is_null:
                    raise ValueError("data {} have nan".format(data_definition))

            # write data
            with SimpleTimer("[{}] Writing generated data {} to hdf5 file"
                             .format(type(self).__name__, data_definition),
                             end_in_new_line=False), \
                    warnings.catch_warnings():
                warnings.simplefilter('ignore', NaturalNameWarning)
                if (isinstance(data, pd.DataFrame)
                        and isinstance(data.index, pd.MultiIndex)
                        and isinstance(data.columns, pd.MultiIndex)):
                    hdf_store.put(data_definition.to_json(), data)
                else:
                    hdf_store.put(data_definition.to_json(), data, format='table')

    def bundle(self, data_definition, path, new_key):
        """Copy the data to another HDF5 file with new key."""
        data = self.get(data_definition).value
        data.to_hdf(path, new_key)
        self.close()

    def close(self):
        if self.hdf_store is not None:
            self.hdf_store.close()
            self.hdf_store = None


class MemoryDataHandler(DataHandler):

    def __init__(self):
        self.data = {}

    def can_skip(self, data_definition):
        if data_definition in self.data:
            return True
        return False

    def get(self, data_definition):
        if isinstance(data_definition, DataDefinition):
            return self.data[data_definition]
        return {k: self.data[k] for k in data_definition}

    def write_data(self, data_definition, data):
        self.data[data_definition] = data


class PickleDataHandler(DataHandler):

    def __init__(self, pickle_dir):
        self.pickle_dir = Path(pickle_dir)
        self.pickle_dir.mkdir(parents=True, exist_ok=True)

    def can_skip(self, data_definition):
        data_path = self.pickle_dir / (data_definition.to_json() + ".pkl")
        if data_path.exists():
            return True
        return False

    def get(self, data_definition):
        if isinstance(data_definition, DataDefinition):
            with (self.pickle_dir / (data_definition.to_json() + ".pkl")).open('rb') as fp:
                return cPickle.load(fp)
        data = {}
        for data_def in data_definition:
            with (self.pickle_dir / (data_def.to_json() + ".pkl")).open('rb') as fp:
                data[data_def] = cPickle.load(fp)
        return data

    def write_data(self, data_definition, data):
        pickle_path = self.pickle_dir / (data_definition.to_json() + ".pkl")
        with SimpleTimer("Writing generated data %s to pickle file" % data_definition,
                         end_in_new_line=False), \
                pickle_path.open('wb') as fp:
            cPickle.dump(data, fp, protocol=cPickle.HIGHEST_PROTOCOL)



from __future__ import print_function, division, absolute_import, unicode_literals
import collections
from collections import OrderedDict
import copy
import json

import six


class FrozenDictJSONEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, FrozenDict):
            return obj._dict
        return super(FrozenDictJSONEncoder, self).default(obj)


class FrozenDict(collections.Mapping):
    def __init__(self, *args, **kwargs):
        self._dict = dict(*args, **kwargs)

    def __getitem__(self, key):
        return self._dict[key]

    def __contains__(self, key):
        return key in self._dict

    def __iter__(self):
        return iter(self._dict)

    def __len__(self):
        return len(self._dict)

    def __repr__(self):
        class_name = type(self).__name__
        return "%s(%s)" % (class_name, repr(self._dict))

    def __str__(self):
        return str(self._dict)

    def __hash__(self):
        if hasattr(self, '_hash'):
            return self._hash
        h = 0
        for key_value in six.viewitems(self._dict):
            h ^= hash(key_value)
        self._hash = h
        return h

    def keys(self):
        return six.viewkeys(self._dict)

    def values(self):
        for val in six.viewvalues(self._dict):
            yield copy.deepcopy(val)

    def items(self):
        for key, val in six.viewitems(self._dict):
            yield key, copy.deepcopy(val)

    def copy(self):
        return copy.copy(self)

    def replace(self, *args, **kwargs):
        new_dict = copy.deepcopy(self._dict)
        new_dict.update(*args, **kwargs)
        return type(self)(new_dict)

    def to_json(self):
        return json.dumps(self, cls=FrozenDictJSONEncoder)

    def __lt__(self, other):
        if isinstance(other, FrozenDict):
            return str(self) < str(other)
        return NotImplemented

    def __le__(self, other):
        if isinstance(other, FrozenDict):
            return str(self) <= str(other)
        return NotImplemented

    def __gt__(self, other):
        if isinstance(other, FrozenDict):
            return str(self) > str(other)
        return NotImplemented

    def __ge__(self, other):
        if isinstance(other, FrozenDict):
            return str(self) >= str(other)
        return NotImplemented

    @classmethod
    def recursively_froze(cls, value):
        if (isinstance(value, collections.Mapping)
                and not isinstance(value, collections.Hashable)):
            value = cls._recursively_froze_mapping(value)
        elif (isinstance(value, collections.Sequence)
                and not isinstance(value, collections.Hashable)):
            value = cls._recursively_froze_sequence(value)
        return value

    @classmethod
    def _recursively_froze_sequence(cls, sequence):
        return tuple(cls.recursively_froze(val) for val in sequence)

    @classmethod
    def _recursively_froze_mapping(cls, mapping):
        return SortedFrozenDict((key, cls.recursively_froze(val))
                                for key, val in six.viewitems(mapping))


class OrderedFrozenDict(FrozenDict):
    def __init__(self, *args, **kwargs):
        self._dict = OrderedDict(*args, **kwargs)

    def __str__(self):
        s = ', '.join('%r: %r' % (k, v) for k, v in six.viewitems(self._dict))
        s = '{%s}' % s
        return s


class SortedFrozenDict(OrderedFrozenDict):
    def __init__(self, *args, **kwargs):
        unsorted_dict = dict(*args, **kwargs)
        self._dict = OrderedDict((k, unsorted_dict[k])
                                 for k in sorted(six.viewkeys(unsorted_dict)))



from __future__ import print_function, division, absolute_import, unicode_literals
import inspect
from collections import Counter
from copy import deepcopy
try:
    from inspect import signature
except ImportError:
    from funcsigs import signature

import six
import networkx as nx
from bistiming import SimpleTimer

from .dag import DataGraph, draw_dag
from .bundling import DataBundlerMixin
from .data_handlers import (
    MemoryDataHandler,
    H5pyDataHandler,
    PandasHDFDataHandler,
    PickleDataHandler,
)


def check_result_dict_keys(result_dict, expected_key_set, function_name):
    """Check whether the ``result_dict`` match the ``expected_key_set``.

    Raises
    ------
    ValueError
        Raise if the keys in ``result_dict`` doesn't match ``expected_key_set``.
    """
    result_dict_key_set = set(result_dict.keys())
    if expected_key_set != result_dict_key_set:
        # format the error message
        missing_keys = expected_key_set - result_dict_key_set
        redundant_keys = result_dict_key_set - expected_key_set
        detail_messages = []
        if missing_keys:
            detail_messages.append('missing_keys: %s' % missing_keys)
        if redundant_keys:
            detail_messages.append('redundant_keys: %s' % redundant_keys)

        err_msg = ("The return keys of function %s %s is not expected. (%s)"
                   % (function_name,
                      result_dict_key_set if result_dict_key_set else "{}",
                      ', '.join(detail_messages)))
        raise ValueError(err_msg)


class DataGeneratorType(type):

    def __init__(cls, name, bases, attrs):  # noqa
        # pylint: disable=protected-access
        super(DataGeneratorType, cls).__init__(name, bases, attrs)

        attrs = inspect.getmembers(
            cls, lambda a: hasattr(a, '_dagian_output_configs'))

        dag = DataGraph()
        # build the dynamic DAG
        handler_set = set()
        for function_name, function in attrs:
            handler_set.update(config['handler'] for config in function._dagian_output_configs)

            # requirements
            if hasattr(function, '_dagian_requirements'):
                requirements = function._dagian_requirements
            else:
                requirements = ()

            # parameters
            sig = signature(function)
            parameters = []
            for name, p in sig.parameters.items():
                # We only take p.kind in {POSITIONAL_OR_KEYWORD, KEYWORD_ONLY} with valid name.
                if name in ('self', 'upstream_data'):
                    continue
                elif p.kind in (p.POSITIONAL_ONLY, p.VAR_POSITIONAL):
                    raise ValueError(
                        "generator methods in dagian.DataGenerator only accept keyword arguments")
                elif p.kind != p.VAR_KEYWORD:
                    parameters.append(p)

            dag.add_node(
                function_name,
                parameters=parameters,
                requirements=requirements,
                output_configs=function._dagian_output_configs,
            )

        cls._dag = dag
        cls._handler_set = handler_set


def _run_function(function, data_definitions, kwargs):
    with SimpleTimer("Generating {} using {}"
                     .format(data_definitions, function.__name__),
                     end_in_new_line=False):  # pylint: disable=C0330
        result_dict = function(**kwargs)
    return result_dict


def _check_result_dict_type(result_dict, function_name):
    if not (hasattr(result_dict, 'keys')
            and hasattr(result_dict, '__getitem__')):
        raise ValueError("the return value of mehod {} should have "
                         "keys and __getitem__ methods".format(function_name))


class DataGenerator(six.with_metaclass(DataGeneratorType, DataBundlerMixin)):

    def __init__(self, handlers):
        handler_set = set(six.viewkeys(handlers))
        if handler_set != self._handler_set:
            redundant_handlers_set = handler_set - self._handler_set
            lacked_handlers_set = self._handler_set - handler_set
            raise ValueError('Handler set mismatch. {} redundant and {} lacked.'
                             .format(redundant_handlers_set,
                                     lacked_handlers_set))
        self._handlers = handlers

    def get_handler(self, key):
        handler_name = self._dag.get_handler_name(key)
        handler = self._handlers[handler_name]
        return handler

    @classmethod
    def check_data_definition(cls, data_definition):
        """Check whether a data definition is valid and set the defaut arguments.

        Parameters
        ----------
        data_definition : DataDefinition

        Returns
        -------
        valid_data_definition : DataDefinition
            Data definition with default values filled.
        """
        key_node_attrs_dict = cls._dag._key_node_attrs_dict
        if data_definition.key not in key_node_attrs_dict:
            raise KeyError("Data key %s doesn't exist." % data_definition.key)
        parameters = key_node_attrs_dict[data_definition.key]['parameters']

        # check parameters
        defined_params = set(data_definition.args.keys())
        valid_params = set(param.name for param in parameters)
        required_params = set(param.name for param in parameters
                              if param.default is param.empty)
        if not defined_params.issubset(valid_params):
            unexpected_args = defined_params - valid_params
            raise ValueError("Data definition %s has %d unexpected arguments %s."
                             % (data_definition, len(unexpected_args), unexpected_args))
        if not required_params.issubset(defined_params):
            missing_args = required_params - defined_params
            raise ValueError("Data definition %s misses %d required arguments %s."
                             % (data_definition, len(missing_args), missing_args))

        # fill default args
        valid_args = {param.name: data_definition.args.get(param.name, param.default)
                      for param in parameters}
        valid_data_definition = data_definition.replace(args=valid_args)
        return valid_data_definition

    @classmethod
    def check_data_definitions(cls, data_definitions):
        """``check_data_definition()`` with ``Sequence`` input and output.

        Parameters
        ----------
        data_definitions : Sequence[DataDefinition]

        Returns
        -------
        valid_data_definitions : Sequence[DataDefinition]
            Data definitions with default values filled.
        """
        valid_data_definitions = [cls.check_data_definition(data_def)
                                  for data_def in data_definitions]
        return valid_data_definitions

    def get(self, data_definition):
        data_definition = self.check_data_definition(data_definition)
        handler = self.get_handler(data_definition.key)
        data = handler.get(data_definition)
        return data

    def _dag_prune_can_skip(self, nx_digraph, generation_order):
        for node in reversed(generation_order):
            node_attrs = nx_digraph.node[node]
            key_info_dict = {key: {'handler': self._handlers[config['handler']]}
                             for key, config in six.viewitems(node_attrs['output_configs'])}
            node_attrs['skipped'] = True
            for target_node, edge_attr in nx_digraph.succ[node].items():
                if nx_digraph.node[target_node]['skipped']:
                    edge_attr['skipped_data'] = edge_attr['data_definitions']
                    edge_attr['nonskipped_data'] = set()
                else:
                    required_data_defs = edge_attr['data_definitions']
                    edge_attr['skipped_data'] = set()
                    edge_attr['nonskipped_data'] = set()
                    for required_data_def in required_data_defs:
                        key_info = key_info_dict[required_data_def.key]
                        if 'can_skip' not in key_info:
                            key_info['can_skip'] = key_info['handler'].can_skip(required_data_def)
                        if key_info['can_skip']:
                            edge_attr['skipped_data'].add(required_data_def)
                        else:
                            edge_attr['nonskipped_data'].add(required_data_def)
                    if len(edge_attr['nonskipped_data']) > 0:
                        node_attrs['skipped'] = False

    def build_involved_dag(self, data_definitions):
        # get the nodes and edges that will be considered during the generation
        data_definitions = self.check_data_definitions(data_definitions)
        involved_dag = self._dag.build_directed_graph(data_definitions, root_node_key='generate')
        generation_order = list(nx.topological_sort(involved_dag))[:-1]
        involved_dag.node['generate']['skipped'] = False
        self._dag_prune_can_skip(involved_dag, generation_order)
        return involved_dag, generation_order

    def draw_involved_dag(self, path, data_definitions):
        involved_dag, _ = self.build_involved_dag(data_definitions)
        draw_dag(involved_dag, path)

    def _get_upstream_data(self, dag, data_definitions):
        data = {}
        name_counter = Counter()
        for source_node, edge_attrs in dag.pred[data_definitions].items():
            source_output_configs = dag.nodes[source_node]['output_configs']
            for pred_def in edge_attrs['data_definitions']:
                source_handler_str = source_output_configs[pred_def.key]['handler']
                source_handler = self._handlers[source_handler_str]
                source_data = source_handler.get(pred_def)
                if name_counter[pred_def.name] == 0:
                    data[pred_def.name] = source_data
                elif name_counter[pred_def.name] == 1:
                    data[pred_def.name] = [data[pred_def.name], source_data]
                else:
                    data[pred_def.name].append(source_data)
                name_counter[pred_def.name] += 1
        return data

    def _generate_one(self, dag, data_definitions, func_name, output_configs):
        # prepare kwargs for function
        data = self._get_upstream_data(dag, data_definitions)
        if data:
            function_kwargs = {'upstream_data': data}
        else:
            function_kwargs = {}
        if data_definitions.args:
            function_kwargs.update(deepcopy(data_definitions.args._dict))

        # TODO: add handler-specific arguments
        # handler = self._handlers[handler_key]
        # function_kwargs = handler.get_function_kwargs(
        #     data_definitions=data_definitions,
        #     data=data,
        #     **handler_kwargs
        # )
        function = getattr(self, func_name)
        result_dict = _run_function(function, data_definitions, function_kwargs)
        self.close()

        if result_dict is None:
            result_dict = {}

        # check result_dict
        _check_result_dict_type(result_dict, func_name)
        expected_keys = set(
            key for key, config in six.viewitems(output_configs)
            if (self._handlers[config['handler']]
                .is_return_data_expected(**config['handler_kwargs'])))
        check_result_dict_keys(result_dict, expected_keys, func_name)

        # write data
        for key in sorted(expected_keys):
            config = output_configs[key]
            data_definition = data_definitions.replace(key=key)
            self._handlers[config['handler']].write_data(
                data_definition, result_dict[key], **config['handler_kwargs'])

    def generate(self, data_definitions, dag_output_path=None):
        """
        Parameters
        ----------
        data_definitions: Sequence[DataDefinition]
        """
        involved_dag, generation_order = self.build_involved_dag(data_definitions)
        if dag_output_path is not None:
            draw_dag(involved_dag, dag_output_path)

        # generate data
        for data_definitions in generation_order:
            node_attrs = involved_dag.nodes[data_definitions]
            if node_attrs['skipped']:
                continue
            self._generate_one(
                involved_dag, data_definitions, node_attrs['func_name'],
                node_attrs['output_configs'])

        return involved_dag

    def close(self):
        for handler in six.viewvalues(self._handlers):
            handler.close()

    @classmethod
    def draw_dag(cls, path, data_definitions):
        # pylint: disable=protected-access
        data_definitions = cls.check_data_definitions(data_definitions)
        dag = cls._dag.draw(path, data_definitions, root_node_key='generate')
        if not nx.is_directed_acyclic_graph(dag):
            print("Warning! The graph is not acyclic!")


class FeatureGenerator(DataGenerator):

    def __init__(self, handlers=None, h5py_hdf_path=None, pandas_hdf_path=None,
                 pickle_dir=None):
        if handlers is None:
            handlers = {}
        if 'memory' in self._handler_set and 'memory' not in handlers:
            handlers['memory'] = MemoryDataHandler()
        if 'h5py' in self._handler_set and 'h5py' not in handlers:
            if h5py_hdf_path is None:
                raise ValueError("h5py_hdf_path should be specified "
                                 "when initiating FeatureGenerator.")
            handlers['h5py'] = H5pyDataHandler(h5py_hdf_path)
        if 'pandas_hdf' in self._handler_set and 'pandas_hdf' not in handlers:
            if pandas_hdf_path is None:
                raise ValueError("pandas_hdf_path should be specified "
                                 "when initiating FeatureGenerator.")
            handlers['pandas_hdf'] = PandasHDFDataHandler(pandas_hdf_path)
        if 'pickle' in self._handler_set and 'pickle' not in handlers:
            if pickle_dir is None:
                raise ValueError("pickle_dir should be specified "
                                 "when initiating FeatureGenerator.")
            handlers['pickle'] = PickleDataHandler(pickle_dir)
        super(FeatureGenerator, self).__init__(handlers)



from __future__ import print_function, division, absolute_import, unicode_literals
from os.path import join
from tempfile import mkdtemp
from shutil import rmtree

import h5py
from dagian.tools.dagian_runner import dagian_run_with_configs


def test_generate_lifetime_features():
    test_output_dir = mkdtemp(prefix="dagian_test_output_")
    h5py_hdf_path = join(test_output_dir, "h5py.h5")
    pandas_hdf_path = join(test_output_dir, "pandas.h5")
    pickle_dir = join(test_output_dir, "pickle")
    data_bundles_dir = join(test_output_dir, "data_bundles")

    global_config = {
        'generator_class': 'dagian.tests.lifetime_feature_generator'
                           '.LifetimeFeatureGenerator',
        'data_bundles_dir': data_bundles_dir,
        'generator_kwargs': {
            'h5py_hdf_path': h5py_hdf_path,
            'pandas_hdf_path': pandas_hdf_path,
            'pickle_dir': pickle_dir,
        },
    }

    bundle_config = {
        'name': 'default',
        'structure': {
            'label': 'lifetime',
            'test_filters': [
                'is_in_test_set',
            ],
            'test_dict': {
                'comparison': [
                    'weight',
                    'height',
                    'mem_raw_data',
                    'pd_weight',
                    'pd_height',
                    'pd_raw_data',
                    # 'pd_raw_data_append',
                ],
            },
            'features': [
                'weight',
                'height',
                'mem_raw_data',
                # 'man_raw_data',
                'pd_weight',
                'pd_height',
                'pd_raw_data',
                'BMI',
                {'key': 'division',
                 'args': {'dividend': 'weight'}},
                {'key': 'division_2_divisor',
                 'args': {'dividend': 'weight',
                          'divisor1': 'height',
                          'divisor2': 'height'}},
                {'key': 'division_pd_2_divisor',
                 'args': {'dividend': 'weight',
                          'divisor1': 'height',
                          'divisor2': 'height'}},
                {'key': 'recursive_division',
                 'args': {'dividend': {'key': 'recursive_division',
                                       'args': {'dividend': 'weight',
                                                'divisor': 'height'}},
                          'divisor': 'height'}},
                {'key': 'sequential_division',
                 'args': {'sequence': [{'key': 'weight'}, 'height', {'key': 'height'}]}},
                'nan',
            ],
        },
        'structure_config': {
            'features': {
                'concat': True,
            }
        }
    }

    dagian_run_with_configs(global_config, bundle_config)

    data_bundle_hdf_path = join(data_bundles_dir, bundle_config['name'] + '.h5')
    with h5py.File(data_bundle_hdf_path, "r") as data_bundle_h5f:
        assert set(data_bundle_h5f) == {'features', 'test_filters', 'label', 'test_dict'}
        assert set(data_bundle_h5f['test_filters']) == {'is_in_test_set'}
        assert set(data_bundle_h5f['test_dict']) == {'comparison'}
        assert (set(data_bundle_h5f['test_dict/comparison'])
                == set(bundle_config['structure']['test_dict']['comparison']))
        assert data_bundle_h5f['features'].shape == (6, 15)

    rmtree(test_output_dir)



from __future__ import print_function, division, absolute_import, unicode_literals
import re

from past.builtins import basestring

from .data_definition import RequirementDefinition, Argument


DATA_KEY_PATTERN = re.compile(r'^[_a-zA-Z][_a-zA-Z0-9]*$')


def require(*args, **kwargs):
    if len(args) == 1:
        data_key = args[0]
        data_name = None
    elif len(args) == 2:
        data_key, data_name = args
    else:
        raise ValueError("require() only accepts 1 or 2 positional arguments.")

    if data_name is None:
        if isinstance(data_key, Argument):
            data_name = data_key.parameter
        elif isinstance(data_key, basestring):
            data_name = data_key
        else:
            raise ValueError("Data key can only be str or Argument.")

    def require_decorator(func):
        # pylint: disable=protected-access
        if not hasattr(func, '_dagian_requirements'):
            func._dagian_requirements = []
        func._dagian_requirements.append(RequirementDefinition(data_key, kwargs, data_name))
        return func
    return require_decorator


def will_generate(data_handler, output_keys, **handler_kwargs):
    """
    Parameters
    ----------
    output_keys: Union[List[str], str]
    """
    if isinstance(output_keys, basestring):
        output_keys = (output_keys,)

    def will_generate_decorator(func):
        # pylint: disable=protected-access
        if not hasattr(func, '_dagian_output_configs'):
            func._dagian_output_configs = []
        for output_key in output_keys:
            matched = DATA_KEY_PATTERN.match(output_key)
            if matched is None:
                raise ValueError("output_key %s doesn't match the pattern %s."
                                 % (output_key, DATA_KEY_PATTERN.pattern))
            func._dagian_output_configs.append({
                'handler': data_handler,
                'key': output_key,
                'handler_kwargs': handler_kwargs,
            })
        return func
    return will_generate_decorator



from __future__ import print_function, division, absolute_import, unicode_literals
import sys
import argparse
import collections

import yaml
from pathlib2 import Path

from .config import get_data_generator_from_config
from ..bundling import get_data_definitions_from_structure


def dagian_run_with_configs(global_config, bundle_config, dag_output_path=None,
                            no_bundle=False):
    """Generate feature with configurations.

    global_config (collections.Mapping): global configuration
        generator_class: string
        data_bundles_dir: string
        generator_kwargs: collections.Mapping

    bundle_config (collections.Mapping): bundle configuration
        name: string
        structure: collections.Mapping
    """
    if not isinstance(global_config, collections.Mapping):
        raise ValueError("global_config should be a "
                         "collections.Mapping object.")
    if not isinstance(bundle_config, collections.Mapping):
        raise ValueError("bundle_config should be a "
                         "collections.Mapping object.")
    data_generator = get_data_generator_from_config(global_config)
    data_definitions = get_data_definitions_from_structure(bundle_config['structure'])
    data_generator.generate(data_definitions, dag_output_path)

    if not no_bundle:
        data_bundles_dir = Path(global_config['data_bundles_dir']).expanduser()
        data_bundles_dir.mkdir(parents=True, exist_ok=True)
        bundle_path = data_bundles_dir / (bundle_config['name'] + '.h5')
        data_generator.bundle(
            bundle_config['structure'], data_bundle_hdf_path=bundle_path,
            structure_config=bundle_config['structure_config'])


def dagian_run(argv=sys.argv[1:]):

    parser = argparse.ArgumentParser(
        description="Generate data cache and bundle.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('-g', '--global-config',
                        default=".dagianrc/config.yml",
                        help="the path of the path configuration YAML file")
    parser.add_argument('-b', '--bundle-config',
                        default=".dagianrc/bundle_config.yml",
                        help="the path of the bundle configuration YAML file")
    parser.add_argument('-d', '--dag-output-path', default=None,
                        help="draw the involved subDAG to the provided path")
    parser.add_argument('--no-bundle', action='store_true',
                        help="not generate the data bundle")
    args = parser.parse_args(argv)
    with open(args.global_config) as fp:
        global_config = yaml.safe_load(fp)
    with open(args.bundle_config) as fp:
        bundle_config = yaml.safe_load(fp)
    bundle_config.setdefault('name', Path(args.bundle_config).stem)
    dagian_run_with_configs(global_config, bundle_config, args.dag_output_path, args.no_bundle)



from __future__ import print_function, division, absolute_import, unicode_literals
import os
from past.builtins import basestring

import numpy as np
import pandas as pd
import h5py
import six
from bistiming import SimpleTimer
from tqdm import trange

from .data_definition import DataDefinition


def get_data_definitions_from_list_in_structure(structure):
    data_definitions = []
    for raw_data_def in structure:
        if isinstance(raw_data_def, dict):
            data_definitions.append(
                DataDefinition(raw_data_def['key'], raw_data_def['args']))
        elif isinstance(raw_data_def, basestring):
            data_definitions.append(DataDefinition(raw_data_def))
        else:
            raise TypeError("The bundle structure in list only support "
                            "dict and str, but got {}.".format(structure))
    return data_definitions


def get_data_definitions_from_dict_in_structure(structure):
    if 'key' in structure:
        data_definitions = [DataDefinition(structure['key'], structure['args'])]
    else:
        data_definitions = []
        for _, val in six.viewitems(structure):
            data_definitions.extend(get_data_definitions_from_structure(val))
    return data_definitions


def get_data_definitions_from_structure(structure):
    if isinstance(structure, basestring):
        data_definitions = [DataDefinition(structure)]
    elif isinstance(structure, list):
        data_definitions = get_data_definitions_from_list_in_structure(structure)
    elif isinstance(structure, dict):
        data_definitions = get_data_definitions_from_dict_in_structure(structure)
    else:
        raise TypeError("The bundle structure only support "
                        "dict, list and str, but got {}.".format(structure))
    return data_definitions


class DataBundlerMixin(object):

    def fill_concat_data(self, data_bundle_hdf_path, dset_name, data_definitions,
                         buffer_size=int(1e+9)):
        data_shapes = []
        for data_definition in data_definitions:
            data_shape = self.get(data_definition).shape
            if len(data_shape) == 1:
                data_shape += (1,)
            data_shapes.append(data_shape)
        max_shape_length = max(map(len, data_shapes))
        if max_shape_length > 2:
            raise NotImplementedError("tensor data is not supported yet")

        n_rows = data_shapes[0][0]
        for data_shape in data_shapes:
            if data_shape[0] != n_rows:
                raise ValueError("different number of instances: {} and {}."
                                 .format(data_shapes[0], data_shape))
        n_cols = sum(shape[1] for shape in data_shapes)
        concat_shape = (n_rows, n_cols)

        h5f = h5py.File(data_bundle_hdf_path)
        dset = h5f.create_dataset(dset_name, shape=concat_shape,
                                  dtype=np.float32)

        data_d = 0
        for data_i, (data_definition, data_shape) in enumerate(zip(data_definitions, data_shapes)):
            data = self.get(data_definition)
            if isinstance(data, pd.DataFrame):
                data = data.values
            batch_size = buffer_size // (data.dtype.itemsize * data_shape[1])
            if batch_size == 0:
                print("Warning! buffer_size not enough to fitted by an "
                      "instance. Trying to use more memory.")
                batch_size = 1
            desc = "({}/{}) Filling {}".format(data_i + 1, len(data_definitions), data_definition)
            for batch_start in trange(0, data_shape[0], batch_size, desc=desc):
                batch_end = min(data_shape[0], batch_start + batch_size)
                data_buffer = data[batch_start: batch_end]
                if isinstance(data_buffer, pd.DataFrame):
                    data_buffer.values
                if len(data_buffer.shape) == 1:
                    data_buffer = data_buffer[:, np.newaxis]
                dset[batch_start: batch_end,
                     data_d: data_d + data_shape[1]] = data_buffer

            data_d += data_shape[1]

        h5f.close()

    def _bundle_list_in_structure(
            self, structure, data_bundle_hdf_path, buffer_size, structure_config, dset_name):
        data_definitions = get_data_definitions_from_list_in_structure(structure)
        if structure_config.get('concat', False):
            # write into single dataset
            self.fill_concat_data(data_bundle_hdf_path, dset_name, data_definitions, buffer_size)
        else:
            key_set = set()
            for data_definition in data_definitions:
                if data_definition.key in key_set:
                    raise ValueError("Duplicated key {} in a list structure."
                                     "Use dict structure to distinguish them instead."
                                     .format(data_definition.key))
                key_set.add(data_definition.key)
                self.get_handler(data_definition.key).bundle(
                    data_definition, data_bundle_hdf_path, dset_name + "/" + data_definition.key)

    def _bundle_dict_in_structure(
            self, structure, data_bundle_hdf_path, buffer_size, structure_config, dset_name):
        if 'key' in structure:
            data_definition = DataDefinition(structure['key'], structure['args'])
            self.get_handler(data_definition.key).bundle(
                data_definition, data_bundle_hdf_path, dset_name)
        else:
            for key, val in six.viewitems(structure):
                self._bundle(
                    val, data_bundle_hdf_path, buffer_size,
                    structure_config=structure_config.get(key, {}),
                    dset_name=dset_name + "/" + key)

    def _bundle(
            self, structure, data_bundle_hdf_path, buffer_size, structure_config, dset_name=""):
        if isinstance(structure, basestring) and dset_name != "":
            self.get_handler(structure).bundle(
                DataDefinition(structure), data_bundle_hdf_path, dset_name)
        elif isinstance(structure, list):
            self._bundle_list_in_structure(
                structure, data_bundle_hdf_path, buffer_size, structure_config, dset_name)
        elif isinstance(structure, dict):
            self._bundle_dict_in_structure(
                structure, data_bundle_hdf_path, buffer_size, structure_config, dset_name)
        else:
            raise TypeError("The bundle structure only support "
                            "dict, list and str (except the first layer).")

    def bundle(self, structure, data_bundle_hdf_path, buffer_size=int(1e+9),
               structure_config=None):
        if structure_config is None:
            structure_config = {}

        data_bundle_hdf_path = str(data_bundle_hdf_path)
        if os.path.isfile(data_bundle_hdf_path):
            os.remove(data_bundle_hdf_path)
        with SimpleTimer("Bundling data"):
            self._bundle(structure, data_bundle_hdf_path, buffer_size, structure_config)
        self.close()



from __future__ import print_function, division, absolute_import, unicode_literals


def get_shape_from_pandas_hdf_storer(storer):
    # TODO: when the data has MultiIndex column and the format is 'fixed', the
    #       shape will be None
    if not storer.is_table:
        shape = storer.shape
        if shape is not None:
            shape = tuple(shape)
    else:
        if storer.pandas_type == 'series_table':
            shape = (storer.nrows,)
        else:
            if storer.is_multi_index:
                ncols = storer.ncols - len(storer.levels)
            else:
                ncols = storer.ncols
            shape = (storer.nrows, ncols)
            assert (shape[0] is not None) and (shape[1] is not None)
    return shape


class PandasHDFDataset(object):
    """h5py Dataset-like wrapper for pandas HDFStore."""

    def __init__(self, hdf_store, key):
        self._hdf_store = hdf_store
        self._storer = hdf_store.get_storer(key)
        self.key = key
        self.shape = get_shape_from_pandas_hdf_storer(self._storer)

    @property
    def value(self):
        return self._hdf_store[self.key]

    @property
    def dtype(self):
        return self._hdf_store.select(self.key, start=0, stop=1).values.dtype

    def select(self, *arg, **kwargs):
        return self._hdf_store.select(self.key, *arg, **kwargs)

    def select_column(self, *arg, **kwargs):
        return self._hdf_store.select_column(self.key, *arg, **kwargs)

    def select_as_coordinates(self, *arg, **kwargs):
        return self._hdf_store.select_as_coordinates(self.key, *arg, **kwargs)

    def __getitem__(self, key):
        if isinstance(key, int):
            return self._hdf_store.select(self.key, start=key, stop=key + 1)
        elif isinstance(key, slice):
            if key.step is None:
                return self._hdf_store.select(self.key,
                                              start=key.start, stop=key.stop)
        raise NotImplementedError("Key {} is not supported".format(key))



from __future__ import print_function, division, absolute_import, unicode_literals
import pkg_resources

__all__ = ['tools', 'bundling', 'data_generator', 'data_handlers', 'decorators', 'dag']
__version__ = pkg_resources.get_distribution("dagian").version

from .data_generator import DataGenerator, FeatureGenerator  # noqa: F401
from .data_definition import Argument  # noqa: F401



from __future__ import print_function, division, absolute_import, unicode_literals
from os.path import join
from tempfile import mkdtemp
import unittest
from shutil import rmtree

import pandas as pd
from dagian.data_wrappers.pandas_hdf import get_shape_from_pandas_hdf_storer


class Test(unittest.TestCase):
    def setUp(self):
        self.test_output_dir = mkdtemp(prefix="dagian_test_output_")
        pandas_hdf_path = join(self.test_output_dir, "pandas.h5")
        self.hdf_store = pd.HDFStore(pandas_hdf_path)

    def tearDown(self):
        self.hdf_store.close()
        rmtree(self.test_output_dir)

    def test_get_shape_from_pandas_hdf_storer_df(self):
        idx = [1, 2, 3, 5, 4]
        col = [10, 9, 6, 7]
        df = pd.DataFrame(0, index=idx, columns=col)
        self.hdf_store['test'] = df
        shape = get_shape_from_pandas_hdf_storer(
            self.hdf_store.get_storer('test'))
        assert shape == (5, 4)

    def test_get_shape_from_pandas_hdf_storer_df_table(self):
        idx = [1, 2, 3, 5, 4]
        col = [10, 9, 6, 7]
        df = pd.DataFrame(0, index=idx, columns=col)
        self.hdf_store.put('test', df, format='table')
        shape = get_shape_from_pandas_hdf_storer(
            self.hdf_store.get_storer('test'))
        assert shape == (5, 4)

    def test_get_shape_from_pandas_hdf_storer_df_m_idx(self):
        idx = pd.MultiIndex.from_product([[0, 1], [0, 1, 2]])
        col = [10, 9, 6, 7]
        df = pd.DataFrame(0, index=idx, columns=col)
        self.hdf_store['test'] = df
        shape = get_shape_from_pandas_hdf_storer(
            self.hdf_store.get_storer('test'))
        assert shape == (6, 4)

    def test_get_shape_from_pandas_hdf_storer_df_m_idx_table(self):
        idx = pd.MultiIndex.from_product([[0, 1], [0, 1, 2]])
        col = [10, 9, 6, 7]
        df = pd.DataFrame(0, index=idx, columns=col)
        self.hdf_store.put('test', df, format='table')
        shape = get_shape_from_pandas_hdf_storer(
            self.hdf_store.get_storer('test'))
        assert shape == (6, 4)

    def test_get_shape_from_pandas_hdf_storer_df_m_col(self):
        idx = [10, 9, 6, 7]
        col = pd.MultiIndex.from_product([[0, 1], [0, 1, 2]])
        df = pd.DataFrame(0, index=idx, columns=col)
        self.hdf_store['test'] = df
        shape = get_shape_from_pandas_hdf_storer(
            self.hdf_store.get_storer('test'))
        # TODO: change to (4, 6)
        assert shape is None

    def test_get_shape_from_pandas_hdf_storer_df_m_col_table(self):
        idx = [10, 9, 6, 7]
        col = pd.MultiIndex.from_product([[0, 1], [0, 1, 2]])
        df = pd.DataFrame(0, index=idx, columns=col)
        self.hdf_store.put('test', df, format='table')
        shape = get_shape_from_pandas_hdf_storer(
            self.hdf_store.get_storer('test'))
        assert shape == (4, 6)

    def test_get_shape_from_pandas_hdf_storer_df_m_idx_m_col(self):
        idx = pd.MultiIndex.from_product([[0, 1], [0, 1, 2]])
        col = pd.MultiIndex.from_product([[0, 1], [0, 1]])
        df = pd.DataFrame(0, index=idx, columns=col)
        self.hdf_store['test'] = df
        shape = get_shape_from_pandas_hdf_storer(
            self.hdf_store.get_storer('test'))
        # TODO: change to (6, 4)
        assert shape is None

    def test_get_shape_from_pandas_hdf_storer_s(self):
        idx = [0, 2, 1, 4, 3]
        s = pd.Series(0, index=idx)
        self.hdf_store['test'] = s
        shape = get_shape_from_pandas_hdf_storer(
            self.hdf_store.get_storer('test'))
        assert shape == (5,)

    def test_get_shape_from_pandas_hdf_storer_s_table(self):
        idx = [0, 2, 1, 4, 3]
        s = pd.Series(0, index=idx)
        self.hdf_store.put('test', s, format='table')
        shape = get_shape_from_pandas_hdf_storer(
            self.hdf_store.get_storer('test'))
        assert shape == (5,)

    def test_get_shape_from_pandas_hdf_storer_s_m_idx(self):
        idx = pd.MultiIndex.from_product([[0, 1], [0, 1, 2]])
        s = pd.Series(0, index=idx)
        self.hdf_store['test'] = s
        shape = get_shape_from_pandas_hdf_storer(
            self.hdf_store.get_storer('test'))
        assert shape == (6,)

    def test_get_shape_from_pandas_hdf_storer_s_m_idx_table(self):
        idx = pd.MultiIndex.from_product([[0, 1], [0, 1, 2]])
        s = pd.Series(0, index=idx)
        self.hdf_store.put('test', s, format='table')
        shape = get_shape_from_pandas_hdf_storer(
            self.hdf_store.get_storer('test'))
        assert shape == (6,)
